[<Microsoft.FSharp.Core.AutoOpen>]
module Degenz.Types

open System
open Degenz

[<Measure>]
type mins

[<Measure>]
type GBT

type Range = { Min : int ; Max : int }

module Range =
    let normalized = { Min = 0 ; Max = 100 }
    let create min max = { Min = min ; Max = max }
    let constrain value range = if value < range.Min then range.Min elif value > range.Max then range.Max else value

type ItemId =
    | Virus = 0
    | RemoteAccess = 1
    | Worm = 2
    | Firewall = 6
    | Encryption = 7
    | Cypher = 8
    | ProteinPowder = 12
    | ToroLoco = 13
    | Cigs = 14
    | MoonPie = 15

type StatId =
    | Strength = 0
    | Focus = 1
    | Charisma = 2
    | Luck = 3

type StatConfig = {
    Id : StatId
    BaseDecayRate : float
    BaseRange : Range
}

type PlayerStat = {
    Id : StatId
    Amount : int
    ModRange : Range
    LastRead : DateTime
}
with static member empty = { Id = StatId.Strength ; Amount = 0 ; ModRange = Range.normalized ; LastRead = DateTime.UtcNow}

type Stats = {
    Strength : PlayerStat
    Focus : PlayerStat
    Luck : PlayerStat
    Charisma : PlayerStat
}
with static member empty = { Strength = PlayerStat.empty ; Focus = PlayerStat.empty ; Luck = PlayerStat.empty ; Charisma = PlayerStat.empty }

type HackResult =
    | Strong
    | Weak

type DiscordPlayer = { Id: uint64; Name: string }
    with static member empty = { Id = 0uL ; Name = "None" }

type HackEvent = {
    IsInstigator : bool
    Adversary : DiscordPlayer
    Success : bool
    HackId : int
}

type PlayerEventType =
    | Hacking of HackEvent
    | Shielding of shieldId : int
    | Stealing of instigator : bool * adversary : DiscordPlayer
    | Imprison

type PlayerEvent =
    { Type : PlayerEventType
      Cooldown : int<mins>
      Timestamp : DateTime }

type HackItem = {
    Id : int
    Name : string
    Price : int<GBT>
    Power : int
    Class : int
    Cooldown : int<mins>
}

type ShieldItem = {
    Id : int
    Name : string
    Price : int<GBT>
    Class : int
    Cooldown : int<mins>
}

type FoodItem = {
    Id : int
    Name : string
    Price : int<GBT>
    TargetStat : StatId
    BoostAmount : int
}

type AccessoryItem = {
    Id : int
    Name : string
    Price : int<GBT>
    TargetStat : StatId
    FloorBoost : int
    CeilBoost : int
}

[<RequireQualifiedAccess>]
type ItemType =
    | Hack
    | Shield
    | Food
    | Accessory

type ItemDetails =
    | Hack of power : int * hackClass : int * cooldown : int<mins>
    | Shield of shieldClass : int * cooldown : int<mins>
    | Food of targetStat : StatId * boostAmount : int
    | Accessory of targetStat : StatId * floorBoost : int * ceilBoost : int
and Item = {
    Id : int
    Name : string
    Price : int<GBT>
//    Type : ItemType
    Details : ItemDetails
}
and Inventory = Item list
and PlayerData = {
    DiscordId : uint64
    Name : string
    Inventory : Inventory
    Events : PlayerEvent list
    Stats : Stats
    Bank : int<GBT>
}
//        Achievements : string array
//        XP : int
with member this.toDiscordPlayer = { Id = this.DiscordId ; Name = this.Name }
     static member empty =
         { DiscordId = 0uL
           Name = "None"
           Inventory = []
           Events = []
           Stats = Stats.empty
//               Achievements = [||]
//               XP = 0
           Bank = 0<GBT> }
